/*
 *  fr.c
 *
 * Start with ppp.c
 * $Id: fr.c,v 1.22 1995/08/05 06:09:53 jfree Exp jfree $
 */
#include "s502.h"

int fr_debug = 2;

int init_module(void);
void cleanup_module(void);
char kernel_version[] = UTS_RELEASE;

static int fr_init(struct device *);
static int fr_open(struct device *);
static void checkCompletionStatus(fr_t *);
static int issueCmd(fr_t *, u_char, u_short);
static void checkReturnCode(fr_t *fr);
static void fr_poll(unsigned long);
static void enableComm(fr_t *fr);
static void infoRead(fr_t *fr);
static int txSend(fr_t *fr);
static int fr_xmit(struct sk_buff *, struct device *);
static void infoWrite(fr_t *fr);
static unsigned short fr_type_trans (struct sk_buff *skb, struct device *dev);
static int fr_header(unsigned char *buff, struct device *dev,
			unsigned short type, void *daddr, void *saddr,
			unsigned len, struct sk_buff *skb);
static int fr_rebuild_header(void *buff, struct device *dev,
				unsigned long raddr, struct sk_buff *skb);
static struct enet_statistics *fr_get_stats (struct device *dev);
static int fr_close(struct device *);

static fr_t	*FrDevs = NULL;
static char	*devName = "fr0";

int
init_module(void)
{	int err;
	struct device	*dev;

/* kmalloc,,, count, register indexed */
	dev = (struct device *) kmalloc(sizeof(struct device), GFP_KERNEL);
	if(!dev)
		return(-ENOMEM);
	memset(dev, 0, sizeof(struct device));
	dev->name = devName;
	dev->init = fr_init;
printk("register_netdev = 0x%X\n", (unsigned)register_netdev);
printk("register_netdevice_notifier = 0x%X\n", (unsigned)register_netdevice_notifier);
	err = register_netdev(dev);
	if (err == -EEXIST) 
		printk("fr: already present. Module not loaded.\n");
	return err;
}

/* called at module load time for each fr device */
int
fr_init(struct device *dev)
{	fr_t	*fr;
	int	i;

	printk("fr: version %s\n", S502_VERSION);

	/* Allocate and initialize struct netstats
		and fr specific stuff
	*/
	fr = (fr_t *) kmalloc(sizeof(fr_t), GFP_KERNEL);
	if (!fr) {
                printk("fr_t alloc failed\n");
                return(-ENOMEM);
	}
	if (!(fr->skq = (struct sk_buff_head *)
			kmalloc(sizeof(struct sk_buff_head), GFP_KERNEL))) {
		printk("sk_buff_head alloc failed\n");
		kfree(fr);
		return(-ENOMEM);
	}
	skb_queue_head_init(fr->skq);
	fr->next = FrDevs;
	fr->shm = (fr_shm_t *)0xd0000;
	FrDevs = fr;

	fr->dev = dev;
	fr->timer.function = fr_poll;

	/* device INFO */
	dev->priv = fr;
	dev->type            = ARPHRD_FR; /* 0, 1 ??? */
	dev->mtu             = FR_MTU;
	dev->open            = fr_open;
	dev->stop            = fr_close;
	dev->do_ioctl        = NULL;	/* ??? */
	dev->addr_len        = 0;
	dev->get_stats       = fr_get_stats;
	dev->hard_header     = fr_header;
	dev->type_trans      = fr_type_trans;
	dev->rebuild_header  = fr_rebuild_header;
	dev->hard_start_xmit = fr_xmit;
	dev->hard_header_len = PADDED_PID_LENGTH;

	for (i = 0; i < DEV_NUMBUFFS; i++) {
		/* dev->buffs[i] = NULL; ??? */
		skb_queue_head_init(&dev->buffs[i]);
	}

	/* New-style flags */
	dev->family     = AF_INET;
	dev->flags	= 0;
	dev->pa_addr    = 0;
	dev->pa_brdaddr = 0;
	dev->pa_mask    = 0;
	dev->pa_alen    = sizeof(unsigned long);

printk("fr_init: done\n");
	return 0;
}

/* Called when fr interface goes "up" as a result of ifconfig on this
 * interface.  Here this just means we begin passing packets.
 */
static int
fr_open(struct device *dev)
{	fr_t *fr = dev->priv;

	/* reset POINTOPOINT every time, since dev_close zaps it! */
	dev->flags |= IFF_POINTOPOINT;

#ifdef TODO
	/* allocate rx, tx buffers ??? */
	(r)mem_[start,end]
	dev: tbusy, interrupt, start
#endif

	PRINTKN (2,(KERN_INFO "fr: channel %s going up!\n",
		dev->name));

	init_timer(&(fr->timer));
	fr->timer.data = (unsigned long)dev;

	if(issueCmd(fr, ENABLE_COMMUNICATIONS, 0))
	 	printk("issueCmd(ENABLE) failed !\n");
	/* Now we need to poll for fr->shm->retcode changing to 0x13, with
	 * data[00,01] = dlci, data[2]:1 meaning the dlci is now active.
	 */

#ifdef DEBUG
	printk("mbox @ 0x%X, cmd @ 0x%X, dlci @ 0x%X, data @ 0x%X, rx indicator @ 0x%X\n",
		fr->shm, &(fr->shm->cmd), &(fr->shm->dlci), &(fr->shm->data),
		&(fr->shm->rxFramesAvail));
	printk("dlci = %d\n", fr->shm->dlci);
#endif /* def DEBUG */

	MOD_INC_USE_COUNT;
	fr_poll((unsigned long)dev);
	return 0;
}

static int
issueCmd(fr_t *fr, u_char cmd, u_short len)
{
	/* cmd, bufLen, data */
	if(!cmd) return(0); /* use assert ??? */
	PRINTKN(4,("@ticks 0x%lX, issueCmd(0x%X, 0x%X)\n", jiffies, cmd, len));

	cli();
	if(!fr->shm->opPending) { /* use assert() , make void ??? */
		fr->cmdPending = cmd;
		fr->shm->cmd = cmd;
		fr->shm->retCode = 0;
		fr->shm->dlci = 16;
		fr->shm->bufLen = len;
		++fr->shm->opPending;
		sti();
		return(0);
	} else {
		sti();
		return(1);
	}
	/* NOTREACHED */
}

/* No command pending, but retCode is set.  Check it out.
 */
static void
checkReturnCode(fr_t *fr)
{
	switch(fr->shm->retCode) {
	case MODEM_FAILURE:
		printk("Modem failure ??\n");
		break;
	case CHANNEL_DOWNED:
		printk("Channel has gone inoperative ??\n");
		break;
	case CHANNEL_UPPED:
		printk("Channel has become operative ??\n");
		break;
	case DLCI_CHANGE:
		printk("DLCI %d changed status to 0x%X ??\n",
			*((u_short *)(fr->shm->data)), fr->shm->data[2]);
		break;
	case DLCI_CONFIG_ERR:
		printk("Status report included non-configured DLCI ??\n");
		break;
	case INVALID_COMMAND:
		printk("Invalid FR command ??\n");
		break;
	default:
		printk("Way funky FR return code 0x%X??\n", fr->shm->retCode);
		break;
	}
	fr->shm->retCode = 0;
	return;
}

static void
checkCompletionStatus(fr_t *fr)
{
#ifdef DEBUG
	printk("chkCmpltStatus: cmd=0x%X, rcode=0x%X, buflen=%d, dlci=0x%X\n",
 		fr->shm->cmd, fr->shm->retCode, fr->shm->bufLen, fr->shm->dlci);
	printk("tx=0x%X rx=0x%X modem=0x%X\n", fr->shm->txSpaceAvail,
		fr->shm->rxFramesAvail, fr->shm->modemStatus);
	printk("appRxDrop=%d intRxDrop=%d\n",
		fr->shm->appRxDrop, fr->shm->intRxDrop);
#endif /* def DEBUG */
	switch(fr->shm->cmd) {
	case INFORMATION_WRITE:
		infoWrite(fr);
		break;
	case INFORMATION_READ:
		infoRead(fr);
		break;
	case ENABLE_COMMUNICATIONS:
		enableComm(fr);
		break;
	default:
		printk("Unhandled cmd 0x%X!\n", fr->shm->cmd);
		break;
	}
	fr->cmdPending = 0;
	fr->shm->cmd = 0;
	return;
}

/* The board is not interrupt driven - we create
 * our own interrupt chain to poll for status.
 */
static void
fr_poll(unsigned long dev)
{	fr_t	*fr = (fr_t *)(((struct device *)dev)->priv);

	if(!fr->shm->opPending) {
		if(fr->cmdPending)
			/* something just finished happening */
			checkCompletionStatus(fr);
		if(fr->shm->retCode)
			checkReturnCode(fr);
		if(fr->txQueued && fr->shm->txSpaceAvail)
			txSend(fr);
		else
			if(fr->shm->rxFramesAvail)
				if(issueCmd(fr, INFORMATION_READ, 0))
					printk("fr_poll: rx avail, can't read!\n");
	}
	/* check modem status ?? */
	/* Re-arm */
	fr->timer.expires = 1;
	add_timer(&(fr->timer));

	return;
}

static void
enableComm(fr_t *fr)
{
	switch(fr->shm->retCode) {
	case OP_SUCCESS:
		printk("S502 communications enabled\n");
		break;
	case DLCI_NOT_CONFIG:
		printk("enableComm() - DLCI configuration not set!\n");
		break;
	default:
		printk("enableComm(0x%X): unknown result 0x%X !?!?\n",
			fr->shm->retCode, fr->shm->retCode);
		break;
	}
}

static void
infoWrite(fr_t *fr)
{
	switch(fr->shm->retCode) {
	case OP_SUCCESS:
		fr->dev->tbusy = 0;
		PRINTKN(4,("infoWrite: tx sent OK\n"));
		break;
	case NOT_ENABLED:
		printk("infoWrite(): NOT_ENABLED\n");
		break;
	case CHANNEL_INOPERATIVE:
		printk("infoWrite(): CHANNEL_INOPERATIVE\n");
		break;
	case DLCI_INACTIVE:
		printk("infoWrite(): DLCI_INACTIVE\n");
		break;
	case DLCI_INVALID:
		printk("infoWrite(): DLCI_INVALID\n");
		break;
	case MAX_LEN_EXCEEDED:
		printk("infoWrite(): MAX_LEN_EXCEEDED\n");
		break;
	case TC_LEN_OVERFLOW:
		printk("infoWrite(): TC_LEN_OVERFLOW\n");
		break;
	case CIR_OVERFLOW:
		printk("infoWrite(): CIR_OVERFLOW\n");
		break;
	case NO_TX_BUFS:
		printk("infoWrite(): NO_TX_BUFS\n");
		break;
	default:
		printk("infoWrite(): unkown result 0x%X\n", fr->shm->retCode);
		break;
	}
	return;
}

static void
infoRead(fr_t *fr)
{
	switch(fr->shm->retCode) {
	case OP_SUCCESS: {
		int	dataLen, pidLen;
		struct sk_buff	*skb;

		PRINTKN(4,("%d bytes received at 0x%lX ticks on dlci %d, bits=0x%X\n",
			fr->shm->bufLen, jiffies, fr->shm->dlci, fr->shm->bits));
#ifdef DEBUG
		{	int i;
			for(i=2; i<26; i++)
				printk("%02X ", fr->shm->data[i]);
		}
		printk("\n");
#endif /* def DEBUG */

#ifdef TODO
	check rx pkt size for runts/giants
	++ rx in stats
#endif
		/* See rfc1490 for NLPID formatting info */
		pidLen = fr->shm->data[1] ? NO_PAD_PID_LENGTH
						: PADDED_PID_LENGTH;
		dataLen = fr->shm->bufLen - pidLen;
		skb = alloc_skb(dataLen + PADDED_PID_LENGTH, GFP_ATOMIC);
		if (skb) {
			skb->len = dataLen + PADDED_PID_LENGTH;
			skb->dev = fr->dev;
			skb->free = 1;
			memcpy(skb->data, fr->shm->data, pidLen);
			memcpy(&(skb->data[PADDED_PID_LENGTH]),
				&(fr->shm->data[pidLen]), dataLen);
			fr->dev->last_rx = jiffies;
			netif_rx(skb);
		} else {
			printk("Unable to allocate skb for FR rx!\n");
		}
	} break;
	case NOT_ENABLED:
		printk("infoRead(): NOT_ENABLED\n");
		break;
	case CHANNEL_INOPERATIVE:
		printk("infoRead(): CHANNEL_INOPERATIVE\n");
		break;
	case NO_FRAMES_QUEUED:
		printk("infoRead(): NO_FRAMES_QUEUED\n");
		break;
	default:
		printk("infoRead(): Unknown result=0x%X\n", fr->shm->retCode);
		break;
	}
	return;
}

static int
txSend(fr_t *fr)
{	u_short	skip;
	struct sk_buff	*skb = NULL;

	if((skb = skb_dequeue(fr->skq)) != NULL) {
		PRINTKN(4, ("txSend: tx skb found - start send\n"));
		fr->dev->trans_start = jiffies;

		/* For now, assume IP.  Later, key off skb->pkt_class */
		skip = skb->data[1] ? 6 : 0;
		/*
		fr->shm->data[0] = 0x03;
		fr->shm->data[1] = 0xCC;
		skip = 2;
		*/

		memcpy(fr->shm->data, &(skb->data[skip]), skb->len - skip);
		if(issueCmd(fr, INFORMATION_WRITE, skb->len - skip))
				printk("txSend: tx avail, can't send!\n");
		fr->dev->tbusy = 1;
		dev_kfree_skb(skb, FREE_WRITE);
		--(fr->txQueued);
	} else
		fr->txQueued = 0;
	return((int)skb);
#ifdef CRUFT
	len = ntohs( ((struct iphdr *)(skb->data)) -> tot_len );
	/* get length from IP header as per Alan Cox bugfix for slip.c */
	if (skb->len < sizeof(struct iphdr)) {
		PRINTKN(0,(KERN_ERR "fr_xmit: given runt packet, ignoring\n"));
		return 1;
	}
	dev->tbusy = 0;
	/* mark_bh (INET_BH); */
	fr->stats.sbusy++;
#endif /* def CRUFT */
}

/*  This routine accepts packets from the network
 *  layer and queues them for delivery.
 */
int
fr_xmit(struct sk_buff *skb, struct device *dev)
{	fr_t	*fr = (fr_t *)(dev->priv);

	PRINTKN(4,(KERN_DEBUG "fr_xmit()\n"));
	if (!(dev->flags & IFF_UP)) {
		PRINTKN(3,("fr_xmit: pkt rejected on downed %s interface!\n",
			dev->name));
		return 1;
	}
	if (skb == NULL) {
		PRINTKN(3,(KERN_WARNING "fr_xmit: null packet!\n"));
		dev_tint(dev);
		return 0;
	}
	skb_queue_tail(fr->skq, skb);
	++fr->txQueued;
	return 0;
}

/* called on ifconfig down ??
 */
static int
fr_close(struct device *dev)
{	fr_t	*fr = (fr_t *)(dev->priv);

	PRINTKN (2,(KERN_INFO "fr: channel %s closing.\n", dev->name));
#ifdef TODO
	dev struct members
	flush buffers
	disable card
	free buffers
	destroy interrupt chain, del_timer(timer)
	init card
#endif
	del_timer(&(fr->timer));
	fr->shm->cmd = DISABLE_COMMUNICATIONS;
	fr->shm->opPending = 1;
	MOD_DEC_USE_COUNT;
/* fr_t still allocated */
	return(0);
}

/* frame type ID.  key off skb->pkt_class */
static unsigned short
fr_type_trans(struct sk_buff *skb, struct device *dev)
{	fr_hdr_t	*frHdr = (fr_hdr_t *)(skb->data);

	PRINTKN(4,(KERN_DEBUG "fr_type_trans()\n"));
	switch(frHdr->pidPad) {
	case 0xCC:
		return(htons(ETH_P_IP));
		break;
	default:
		printk("fr_type_trans(): unkown NLPID 0x%X !\n", frHdr->pidPad);
		return(htons(ETH_P_IP));
		break;
	case 0x00: /* Padding - SNAP header follows */
		switch(frHdr->NLPID) {
		case 0x80: /* SNAP */
			switch(frHdr->PID) {
			case 0x0800:
				return(htons(ETH_P_IP));
				break;
			case 0x8137:
				return(htons(ETH_P_IPX));
				break;
			default:
				printk("fr_type_trans(): unkown PID 0x%X !\n",
					frHdr->PID);
				return(htons(ETH_P_IP));
				break;
			}
		default:
			printk("fr_type_trans(): unknown NLPID 0x%X !\n",
				frHdr->NLPID);
			return(htons(ETH_P_IP));
			break;
		}
		/*NOTREACHED*/
		printk("fr_type_trans(): DANGER ! Shouldn't get here !\n");
		return(htons(ETH_P_IP));
		break;
	}
	/*NOTREACHED*/
	printk("fr_type_trans(): DANGER ! Shouldn't get here !\n");
	return(htons(ETH_P_IP));
}

static int
fr_header(unsigned char *buff, struct device *dev, unsigned short type,
	   void *daddr, void *saddr, unsigned len, struct sk_buff *skb)
{
	PRINTKN(4,(KERN_DEBUG "fr_header()\n"));
if(type == ETH_P_IP) {
/* Accomodate different NLPIDs, not just 0xCC (IP) */
 buff[1] = 0xCC;
 buff[6] = 0x03;
 buff[7] = 0xCC;
} else {
printk("fr_header(): non-IP !\n");
 buff[0] = 0x03;
 buff[1] = 0x00;
/* fill in SNAP info */
}
	return(dev->hard_header_len);
}

static int
fr_rebuild_header(void *buff, struct device *dev, unsigned long raddr,
		   struct sk_buff *skb)
{
printk("fr_rebuild_header():\n");
	PRINTKN(4,(KERN_DEBUG "fr_rebuild_header()\n"));
	return(0);
}

static struct enet_statistics *
fr_get_stats (struct device *dev)
{	fr_t *fr = (fr_t *)dev->priv;
	static struct enet_statistics fr_stats;

	PRINTKN (4, (KERN_DEBUG "fr_get_stats called"));
	fr_stats.tx_heartbeat_errors = 7;
	fr->stats.sbusy = 0;

	return &fr_stats;
}

void
cleanup_module(void)
{	fr_t	*fr = FrDevs, *tmp;
	struct sk_buff	*skb;

	if (MOD_IN_USE)  {
		printk("fr: device busy, forced removal!\n");
		return;
	}
	while(fr) {
		tmp = fr->next;
		while((skb = skb_dequeue(fr->skq)) != NULL)
			kfree_skb(skb, FREE_WRITE);
		kfree(fr->skq);
		fr->skq = NULL;
		fr->dev->start = 0;
		unregister_netdev(fr->dev);
		kfree(fr->dev);
		kfree(fr);
		fr = tmp;
	}
	printk("fr: Module unloaded.\n");
	return;
}
